(*
    Prodiams - Programmable Diagrams.
    Copyright (C) 2017 Inspirel <http://www.inspirel.com/>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

(* ::Package:: *)

Needs["Prodiams`"];


BeginPackage["ProdiamsElectronics`"];


ground::usage="ground[name,OptionsPattern[{style->None,comment->None}]] creates a ground symbol, which itself is a point-like structure without any explicit connectors.";


resistor::usage="resistor[name,OptionsPattern[{label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a resistor symbol.\n"<>
"This shape has connectors named A and B.";


capacitor::usage="capacitor[name,OptionsPattern[{polarized\[Rule]False,label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a capacitor symbol for both polarized and non-polarized variants.\n"<>
"This shape has connectors named A and B.";


diode::usage="diode[name,OptionsPattern[{zener\[Rule]False,label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a diode symbol with optional Zener marks.\n"<>
"This shape has connectors named A and B.";


zenerDiode::usage="zenerDiode[name,OptionsPattern[{label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] convenience wrapper for the diode function, creates a Zener diode symbol.\n"<>
"This shape has connectors named A and B.";


voltageSource::usage="voltageSource[name,OptionsPattern[{label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a voltage source symbol.\n"<>
"This shape has connectors named A and B.";


currentSource::usage="currentSource[name,OptionsPattern[{label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a current voltage symbol.\n"<>
"This shape has connectors named A and B.";


signalSource::usage="signalSource[name,OptionsPattern[{label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a generic signal source symbol.\n"<>
"This shape has connectors named A and B.";


bipolarTransistor::usage="bipolarTransistor[name,type,OptionsPattern[{orientation\[Rule]Right,rotation\[Rule]0,label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a bipolar transistor symbol of the given type (1-NPN, 2-PNP), orientation (Left/Right) and rotation, expressed in the number (0-3) of 90\[Degree] units.\n"<>
"This shape has connectors named B, C and E.";


transistorNPN::usage="transistorNPN[name,OptionsPattern[{orientation\[Rule]Right,rotation\[Rule]0,label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates an NPN transistor symbol of the given orientation (Left/Right) and rotation, expressed in the number (0-3) of 90\[Degree] units.\n"<>
"This shape has connectors named B, C and E.";


transistorPNP::usage="transistorPNP[name,OptionsPattern[{orientation\[Rule]Right,rotation\[Rule]0,label\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a PNP transistor symbol of the given orientation (Left/Right) and rotation, expressed in the number (0-3) of 90\[Degree] units.\n"<>
"This shape has connectors named B, C and E.";


Begin["`Private`"];


(* ::Text:: *)
(*Helpers, transform a given base shape (which is a list of graphics primitives) by rotating, translating and connecting it with given starting and ending points:*)


(* transform only graphics primitives like Line, Circle or Rectangle and ignore (preserve without any transformation) others like style directives *)
transformableGraphicsPrimitives={Circle,Line,Parallelogram,Rectangle,Triangle};

rotateAndMove[shape_List,angle_,to_]:=With[
	{
		rotation=RotationTransform[angle],
		translation=TranslationTransform[to]
	},

	Map[
		Function[p,
			If[MemberQ[transformableGraphicsPrimitives,Head[p]],
				TransformedRegion[TransformedRegion[p,rotation],translation],
				p
			]
		],
		shape
	]
]

reflectToLeft[shape_List]:=With[
	{
		reflection=ReflectionTransform[{1,0}]
	},

	Map[
		Function[p,
			If[MemberQ[transformableGraphicsPrimitives,Head[p]],
				TransformedRegion[p,reflection],
				p
			]
		],
		shape
	]
]

connectPointsWithBaseShape[shape_List,shapeWidth_,from_,to_,style_]:=Module[
	{
		xFrom=from[[1]],
		yFrom=from[[2]],
		xTo=to[[1]],
		yTo=to[[2]],
		fromConnectorPoint, (* the part of shape to which "from" point should be connected *)
		toConnectorPoint, (* the part of shape to which "to" point should be connected *)
		distance,center,
		halfWidth=shapeWidth/2,
		dx,dy,angle,
		rotation,translation
	},

	center=Mean[{from,to}];

	dx=xTo-xFrom;
	dy=yTo-yFrom;
	distance=Norm[{dx,dy}];
	angle=If[dx==0,
		If[dy>0,\[Pi]/2,-\[Pi]/2],
		If[dy>=0,
			If[dx>0,
				ArcTan[dy/dx],
				\[Pi]-ArcTan[-dy/dx]
			],
			If[dx>0,
				ArcTan[dy/dx],
				-\[Pi]+ArcTan[dy/dx]
			]
		]
	];

	fromConnectorPoint={center[[1]]-halfWidth/distance*dx,center[[2]]-halfWidth/distance*dy};
	toConnectorPoint={center[[1]]+halfWidth/distance*dx,center[[2]]+halfWidth/distance*dy};
	rotation=RotationTransform[angle];
	translation=TranslationTransform[center];

	Flatten[{
		style[Prodiams`lineStyle],
		rotateAndMove[shape,angle,center],
		style[Prodiams`connectorLineStyle],
		Line[{from,fromConnectorPoint}],
		Line[{toConnectorPoint,to}]
	}]
]


(* ::Text:: *)
(*Common implementations for shapes with similar properties:*)


electronicsLabelOffsets[dx_,dy_]:=With[
	{
		norm=Max[Abs[dx],Abs[dy]]
	},

	{-dx/norm,-dy/norm}
]

electronicsInternalCoordinates[element_]:=With[
	{
		name=element[[1]]
	},

	{
		name["A"]["x"],name["A"]["y"],
		name["B"]["x"],name["B"]["y"]
	}
]

electronicsInternalLayout[element_]:={}

electronicsBoundingBox[element_]:=With[
	{
		name=element[[1]]
	},

	{
		{Min[name["A"]["x"],name["B"]["x"]],Min[name["A"]["y"],name["B"]["y"]]},
		{Max[name["A"]["x"],name["B"]["x"]],Max[name["A"]["y"],name["B"]["y"]]}
	}
]

electronicsSpecificStyle[element_]:=element[[6]]

electronicsDraw[element_,layout_,style_]:=Module[
	{
		name=element[[1]],
		label=element[[2]],
		baseShape=element[[3]],
		baseShapeWidth=element[[4]],
		labelDistance=element[[5]],
		comment=element[[7]],
		xA,yA,
		xB,yB,
		dx,dy,
		xMid,yMid,length,
		dxLabel,dyLabel, (* label displacement from shape center {xMid,yMid} *)
		xLabel,yLabel,
		labelOffsets, (* offsets for style[labelFunction], in the {xOff,yOff} form *)
		drawing
	},

	xA=layout[name["A"]["x"]];
	yA=layout[name["A"]["y"]];
	xB=layout[name["B"]["x"]];
	yB=layout[name["B"]["y"]];

	(* compute the label placement, if any label is specified: *)
	If[label=!=None,
		If[label===Automatic,
			label=ToString[name]
		];

		dx=xB-xA;
		dy=yB-yA;
		xMid=Mean[{xA,xB}];
		yMid=Mean[{yA,yB}];
		length=Norm[{dx,dy}];
		If[xA==xB,
			(* label is on the left of shape, independently on the A-B orientation *)
			dxLabel=-labelDistance;
			dyLabel=0,

			(* label is above and right or left of the shape, depending on dx and dy signs *)
			dxLabel=-(dy/length)*labelDistance;
			dyLabel=(dx/length)*labelDistance
		];

		xLabel=xMid+dxLabel;
		yLabel=yMid+dyLabel;
		labelOffsets=electronicsLabelOffsets[dxLabel,dyLabel];
	];

	(* base shape is a shape in basic orientation and with its center at {0,0} *)
	(* base shape is rotated and moved to its target position, and connectors are added automatically *)

	drawing=Flatten[{
		connectPointsWithBaseShape[baseShape,baseShapeWidth,{xA,yA},{xB,yB},style],

		(* label, if specified *)
		If[label=!=None,
			style[Prodiams`labelFunction][label,{xLabel,yLabel},labelOffsets],
			Nothing
		]
	}];

	(* comment, if specified *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]

electronicsConnector[element_,conName_]:=With[
	{
		name=element[[1]]
	},

	Prodiams`anchorT[name[conName]["x"],name[conName]["y"]]
]


ground[name_,OptionsPattern[{style->None,comment->None}]]:=With[
	{
		groundName=SymbolName[Unevaluated[name]]
	},

	name=groundT[
		groundName["x"],
		groundName["y"],
		OptionValue[style],
		OptionValue[comment]
	]
]

SetAttributes[ground,HoldFirst]

Prodiams`internalCoordinates[g_groundT]:={g[[1]],g[[2]]}

Prodiams`internalLayout[g_groundT]:={}

Prodiams`boundingBox[g_groundT]:={{g[[1]]-1,g[[2]]-2},{g[[1]]+1,g[[2]]}}

Prodiams`specificStyle[g_groundT]:=g[[3]]

Prodiams`draw[g_groundT,layout_,style_]:=Module[
	{
		xName=g[[1]],
		yName=g[[2]],
		comment=g[[4]],
		x,y,
		drawing
	},

	x=layout[xName];
	y=layout[yName];

	drawing=Flatten[{
		style[Prodiams`connectorLineStyle],
		Line[{{x,y},{x,y-2}}],
		style[Prodiams`lineStyle],
		Line[{{x-1.5,y-2},{x+1.5,y-2}}],
		Line[{{x-1,y-2.5},{x+1,y-2.5}}],
		Line[{{x-0.5,y-3},{x+0.5,y-3}}]
	}];

	(* comment, if specified *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]


resistor[name_,OptionsPattern[{label->None,style->None,comment->None}]]:=
	name=resistorT[
		SymbolName[Unevaluated[name]],
		OptionValue[label],
		(* base shape: *)
		{
			With[{h=Sqrt[3]/4},
				Line[{{-2,0},{-(5/3),h},{-1,-h},{-(1/3),h},{1/3,-h},{1,h},{5/3,-h},{2,0}}]
			]
		},
		4, (* base shape width *)
		0.8, (* label distance from shape center *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[resistor,HoldFirst]

Prodiams`internalCoordinates[r_resistorT]:=electronicsInternalCoordinates[r]

Prodiams`internalLayout[r_resistorT]:=electronicsInternalLayout[r]

Prodiams`boundingBox[r_resistorT]:=electronicsBoundingBox[r]

Prodiams`specificStyle[r_resistorT]:=electronicsSpecificStyle[r]

Prodiams`draw[r_resistorT,layout_,style_]:=electronicsDraw[r,layout,style]

Prodiams`connector[r_resistorT,conName_]:=electronicsConnector[r,conName]


capacitor[name_,OptionsPattern[{polarized->False,label->None,style->None,comment->None}]]:=
	name=capacitorT[
		SymbolName[Unevaluated[name]],
		OptionValue[label],
		(* base shape: *)
		{
			Line[{{-0.5,-1.5},{-0.5,1.5}}],
			If[OptionValue[polarized],
				Circle[{4,0},3.5,{\[Pi]-0.45,\[Pi]+0.45}],
				Line[{{0.5,-1.5},{0.5,1.5}}]
			]
		},
		1, (* base shape width *)
		2, (* label distance from shape center *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[capacitor,HoldFirst]

Prodiams`internalCoordinates[c_capacitorT]:=electronicsInternalCoordinates[c]

Prodiams`internalLayout[c_capacitorT]:=electronicsInternalLayout[c]

Prodiams`boundingBox[c_capacitorT]:=electronicsBoundingBox[c]

Prodiams`specificStyle[c_capacitorT]:=electronicsSpecificStyle[c]

Prodiams`draw[c_capacitorT,layout_,style_]:=electronicsDraw[c,layout,style]

Prodiams`connector[c_capacitorT,conName_]:=electronicsConnector[c,conName]


diode[name_,OptionsPattern[{zener->False,label->None,style->None,comment->None}]]:=
	name=diodeT[
		SymbolName[Unevaluated[name]],
		OptionValue[label],
		(* base shape: *)
		{
			Triangle[{{-1,-1},{-1,1},{1,0}}],
			If[OptionValue[zener],
				Line[{{0.5,1},{1,1},{1,-1},{1.5,-1}}],
				Line[{{1,-1},{1,1}}]
			]
		},
		2, (* base shape width *)
		1.5, (* label distance from shape center *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[diode,HoldFirst]

zenerDiode[name_,OptionsPattern[{label->None,style->None,comment->None}]]:=diode[name,zener->True,label->OptionValue[label],style->OptionValue[style],comment->OptionValue[comment]]

SetAttributes[zenerDiode,HoldFirst]

Prodiams`internalCoordinates[d_diodeT]:=electronicsInternalCoordinates[d]

Prodiams`internalLayout[d_diodeT]:=electronicsInternalLayout[d]

Prodiams`boundingBox[d_diodeT]:=electronicsBoundingBox[d]

Prodiams`specificStyle[d_diodeT]:=electronicsSpecificStyle[d]

Prodiams`draw[d_diodeT,layout_,style_]:=electronicsDraw[d,layout,style]

Prodiams`connector[d_diodeT,conName_]:=electronicsConnector[d,conName]


voltageSource[name_,OptionsPattern[{label->None,style->None,comment->None}]]:=
	name=voltageSourceT[
		SymbolName[Unevaluated[name]],
		OptionValue[label],
		(* base shape: *)
		{
			Line[{{-1,1.5},{-1,-1.5}}],
			Line[{{-(1/3),2/3},{-(1/3),-(2/3)}}],
			Line[{{1/3,1.5},{1/3,-1.5}}],
			Line[{{1,2/3},{1,-(2/3)}}]
		},
		2, (* base shape width *)
		2, (* label distance from shape center *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[voltageSource,HoldFirst]

Prodiams`internalCoordinates[vs_voltageSourceT]:=electronicsInternalCoordinates[vs]

Prodiams`internalLayout[vs_voltageSourceT]:=electronicsInternalLayout[vs]

Prodiams`boundingBox[vs_voltageSourceT]:=electronicsBoundingBox[vs]

Prodiams`specificStyle[vs_voltageSourceT]:=electronicsSpecificStyle[vs]

Prodiams`draw[vs_voltageSourceT,layout_,style_]:=electronicsDraw[vs,layout,style]

Prodiams`connector[vs_voltageSourceT,conName_]:=electronicsConnector[vs,conName]


currentSource[name_,OptionsPattern[{label->None,style->None,comment->None}]]:=
	name=currentSourceT[
		SymbolName[Unevaluated[name]],
		OptionValue[label],
		(* base shape: *)
		{
			Circle[{0,0},1.5],
			Line[{{-0.8,0},{0.8,0}}],
			Triangle[{{0.2,0.2},{0.8,0},{0.2,-0.2}}]
		},
		3, (* base shape width *)
		2, (* label distance from shape center *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[currentSource,HoldFirst]

Prodiams`internalCoordinates[cs_currentSourceT]:=electronicsInternalCoordinates[cs]

Prodiams`internalLayout[cs_currentSourceT]:=electronicsInternalLayout[cs]

Prodiams`boundingBox[cs_currentSourceT]:=electronicsBoundingBox[cs]

Prodiams`specificStyle[cs_currentSourceT]:=electronicsSpecificStyle[cs]

Prodiams`draw[cs_currentSourceT,layout_,style_]:=electronicsDraw[cs,layout,style]

Prodiams`connector[cs_currentSourceT,conName_]:=electronicsConnector[cs,conName]


signalSource[name_,OptionsPattern[{label->None,style->None,comment->None}]]:=
	name=signalSourceT[
		SymbolName[Unevaluated[name]],
		OptionValue[label],
		(* base shape: *)
		{
			Circle[{0,0},1.5],
			Circle[{0,0.4},0.4,{-\[Pi]/2,\[Pi]/2}],
			Circle[{0,-0.4},0.4,{\[Pi]/2,3\[Pi]/2}]
		},
		3, (* base shape width *)
		2, (* label distance from shape center *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[signalSource,HoldFirst]

Prodiams`internalCoordinates[ss_signalSourceT]:=electronicsInternalCoordinates[ss]

Prodiams`internalLayout[ss_signalSourceT]:=electronicsInternalLayout[ss]

Prodiams`boundingBox[ss_signalSourceT]:=electronicsBoundingBox[ss]

Prodiams`specificStyle[ss_signalSourceT]:=electronicsSpecificStyle[ss]

Prodiams`draw[ss_signalSourceT,layout_,style_]:=electronicsDraw[ss,layout,style]

Prodiams`connector[ss_signalSourceT,conName_]:=electronicsConnector[ss,conName]


bipolarTransistor[name_,type_,OptionsPattern[{orientation->Right,rotation->0,label->None,style->None,comment->None}]]:=
	name=bipolarTransistorT[
		SymbolName[Unevaluated[name]],
		type,
		OptionValue[orientation],
		OptionValue[rotation],
		OptionValue[label],
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[bipolarTransistor,HoldFirst]

Prodiams`internalCoordinates[t_bipolarTransistorT]:=With[
	{
		name=t[[1]]
	},

	{
		name["B"]["x"],name["B"]["y"],
		name["C"]["x"],name["C"]["y"],
		name["E"]["x"],name["E"]["y"]
	}
]

Prodiams`internalLayout[t_bipolarTransistorT]:=With[
	{
		name=t[[1]],
		orientation=t[[3]],
		rotation=t[[4]]
	},

	If[orientation===Right,
		Switch[rotation,
			0, (* no rotation *)
			{
				name["C"]["x"]==name["B"]["x"]+3,
				name["C"]["y"]==name["B"]["y"]+2,
				name["E"]["x"]==name["B"]["x"]+3,
				name["E"]["y"]==name["B"]["y"]-2
			},
			1, (* rotation \[Pi]/2 *)
			{
				name["C"]["x"]==name["B"]["x"]-2,
				name["C"]["y"]==name["B"]["y"]+3,
				name["E"]["x"]==name["B"]["x"]+2,
				name["E"]["y"]==name["B"]["y"]+3
			},
			2, (* rotation \[Pi] *)
			{
				name["C"]["x"]==name["B"]["x"]-3,
				name["C"]["y"]==name["B"]["y"]-2,
				name["E"]["x"]==name["B"]["x"]-3,
				name["E"]["y"]==name["B"]["y"]+2
			},
			3, (* rotation 3\[Pi]/2 *)
			{
				name["C"]["x"]==name["B"]["x"]+2,
				name["C"]["y"]==name["B"]["y"]-3,
				name["E"]["x"]==name["B"]["x"]-2,
				name["E"]["y"]==name["B"]["y"]-3
			}
		],

		(* orientation Left *)
		Switch[rotation,
			0, (* no rotation *)
			{
				name["C"]["x"]==name["B"]["x"]-3,
				name["C"]["y"]==name["B"]["y"]+2,
				name["E"]["x"]==name["B"]["x"]-3,
				name["E"]["y"]==name["B"]["y"]-2
			},
			1, (* rotation \[Pi]/2 *)
			{
				name["C"]["x"]==name["B"]["x"]-2,
				name["C"]["y"]==name["B"]["y"]-3,
				name["E"]["x"]==name["B"]["x"]+2,
				name["E"]["y"]==name["B"]["y"]-3
			},
			2, (* rotation \[Pi] *)
			{
				name["C"]["x"]==name["B"]["x"]+3,
				name["C"]["y"]==name["B"]["y"]-2,
				name["E"]["x"]==name["B"]["x"]+3,
				name["E"]["y"]==name["B"]["y"]+2
			},
			3, (* rotation 3\[Pi]/2 *)
			{
				name["C"]["x"]==name["B"]["x"]+2,
				name["C"]["y"]==name["B"]["y"]+3,
				name["E"]["x"]==name["B"]["x"]-2,
				name["E"]["y"]==name["B"]["y"]+3
			}
		]
	]
]

Prodiams`boundingBox[t_bipolarTransistorT]:=With[
	{
		name=t[[1]],
		orientation=t[[3]],
		rotation=t[[4]]
	},

	If[orientation===Right,
		Switch[rotation,
			0, (* no rotation *)
			{{name["B"]["x"],name["E"]["y"]},{name["C"]["x"],name["C"]["y"]}},
			1, (* rotation \[Pi]/2 *)
			{{name["C"]["x"],name["B"]["y"]},{name["E"]["x"],name["E"]["y"]}},
			2, (* rotation \[Pi] *)
			{{name["C"]["x"],name["C"]["y"]},{name["B"]["x"],name["E"]["y"]}},
			3, (* rotation 3\[Pi]/2 *)
			{{name["E"]["x"],name["E"]["y"]},{name["C"]["x"],name["B"]["y"]}}
		],

		(* orientation Left *)
		Switch[rotation,
			0, (* no rotation *)
			{{name["E"]["x"],name["E"]["y"]},{name["B"]["x"],name["C"]["y"]}},
			1, (* rotation \[Pi]/2 *)
			{{name["C"]["x"],name["C"]["y"]},{name["E"]["x"],name["B"]["y"]}},
			2, (* rotation \[Pi] *)
			{{name["B"]["x"],name["C"]["y"]},{name["E"]["x"],name["E"]["y"]}},
			3, (* rotation 3\[Pi]/2 *)
			{{name["E"]["x"],name["B"]["y"]},{name["C"]["x"],name["C"]["y"]}}
		]
	]
]

Prodiams`specificStyle[t_bipolarTransistorT]:=t[[6]]

Prodiams`draw[t_bipolarTransistorT,layout_,style_]:=Module[
	{
		name=t[[1]],
		type=t[[2]],
		orientation=t[[3]],
		rotation=t[[4]],
		label=t[[5]],
		comment=t[[7]],
		xB,yB,
		arrowWidth=0.2,
		drawing,labelLocation
	},

	xB=layout[name["B"]["x"]];
	yB=layout[name["B"]["y"]];

	(* base shape has orientation Right and no rotation, with "B" connector at {0,0} *)
	drawing=Flatten[{
		style[Prodiams`lineStyle],
		Line[{{0,0},{1,0}}],
		Rectangle[{1,-1.8},{1.1,1.8}],
		Line[{{1,2/3},{3,2}}],
		Line[{{1,-2/3},{3,-2}}],
		If[type==1,
			(* NPN transistor, arrow pointing outwards *)
			Triangle[{{2+arrowWidth,-4/3+arrowWidth*3/2},{2-arrowWidth,-4/3-arrowWidth*3/2},{3,-2}}],

			(* PNP transistor, arrow pointing inwards *)
			Triangle[{{2+arrowWidth,-4/3+arrowWidth*3/2},{2-arrowWidth,-4/3-arrowWidth*3/2},{1,-2/3}}]
		]
	}];

	(* reflect if necessary (for Left orientation) *)
	If[orientation===Left,
		drawing=reflectToLeft[drawing]
	];

	(* rotate and move the whole shape accordingly *)
	drawing=rotateAndMove[drawing,rotation*\[Pi]/2,{xB,yB}];

	(* add label, if specified *)
	If[label=!=None,
		If[label===Automatic,
			label=ToString[name]
		];

		(* labelLocation is a pair {{x,y},{xOff,yOff}}, decomposed later for labelFunction *)
		labelLocation=If[orientation===Right,
			Switch[rotation,
				0, (* no rotation *)
				{{xB+3.5,yB},{-1,0}},
				1, (* rotation \[Pi]/2 *)
				{{xB,yB+3.5},{0,-1}},
				2, (* rotation \[Pi] *)
				{{xB-3.5,yB},{1,0}},
				3, (* rotation 3\[Pi]/2 *)
				{{xB,yB-3.5},{0,1}}
			],

			(* orientation Left *)
			Switch[rotation,
				0, (* no rotation *)
				{{xB-3.5,yB},{1,0}},
				1, (* rotation \[Pi]/2 *)
				{{xB,yB-3.5},{0,1}},
				2, (* rotation \[Pi] *)
				{{xB+3.5,yB},{-1,0}},
				3, (* rotation 3\[Pi]/2 *)
				{{xB,yB+3.5},{0,-1}}
			]
		];

		AppendTo[drawing,
			style[Prodiams`labelFunction][label,labelLocation[[1]],labelLocation[[2]]]
		]
	];

	(* comment, if specified *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]

Prodiams`connector[t_bipolarTransistorT,conName_]:=With[
	{
		name=t[[1]]
	},

	Prodiams`anchorT[name[conName]["x"],name[conName]["y"]]
]


transistorNPN[name_,OptionsPattern[{orientation->Right,rotation->0,label->None,style->None,comment->None}]]:=
	bipolarTransistor[name,1,
		orientation->OptionValue[orientation],
		rotation->OptionValue[rotation],
		label->OptionValue[label],
		style->OptionValue[style],
		comment->OptionValue[comment]
	]

SetAttributes[transistorNPN,HoldFirst]


transistorPNP[name_,OptionsPattern[{orientation->Right,rotation->0,label->None,style->None,comment->None}]]:=
	bipolarTransistor[name,2,
		orientation->OptionValue[orientation],
		rotation->OptionValue[rotation],
		label->OptionValue[label],
		style->OptionValue[style],
		comment->OptionValue[comment]
	]

SetAttributes[transistorPNP,HoldFirst]


End[];


EndPackage[];
