(*
    Prodiams - Programmable Diagrams.
    Copyright (C) 2017 Inspirel <http://www.inspirel.com/>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

(* ::Package:: *)

Needs["Prodiams`"];


BeginPackage["ProdiamsLogic`"];


logicNOT::usage="logicNOT[name,inName,outName,OptionsPattern[{labels\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a NOT gate with named connectors and optional labels.";


logicAND::usage="logicAND[name,inNames,outName,OptionsPattern[{labels\[Rule]None,outputNegated\[Rule]False,style\[Rule]None,comment\[Rule]None}]] creates an AND gate with named connectors and optional labels.";


logicNAND::usage="logicNAND[name,inNames,outName,OptionsPattern[{labels\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a NAND gate with named connectors and optional labels.";


logicOR::usage="logicOR[name,inNames,outName,OptionsPattern[{labels\[Rule]None,outputNegated\[Rule]False,inputsExclusive\[Rule]False,style\[Rule]None,comment\[Rule]None}]] creates an OR gate with named connectors and optional labels.";


logicNOR::usage="logicNOR[name,inNames,outName,OptionsPattern[{labels\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a NOR gate with named connectors and optional labels.";


logicXOR::usage="logicXOR[name,inNames,outName,OptionsPattern[{labels\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a XOR gate with named connectors and optional labels.";


logicXNOR::usage="logicXNOR[name,inNames,outName,OptionsPattern[{labels\[Rule]None,style\[Rule]None,comment\[Rule]None}]] creates a XNOR gate with named connectors and optional labels.";


logicBlock::usage="logicBlock[name,label,inNames,outNames,OptionsPattern[{width\[Rule]16,connectorLabels\[Rule]Automatic,style\[Rule]None,comment\[Rule]None}]] creates a generic block with the given label, number of inputs and outputs.";


logicDiagram::usage="logicDiagram[expr_,OptionsPattern[{style\[Rule]None,debug\[Rule]False}]] creates a best-effort diagram from the given logical expression.";


(* ::Text:: *)
(*Note : the context is switched here to the "parent" package so that functions that operate on elements created here and called from the main engine can be properly resolved:*)


Begin["`Private`"];


(* ::Text:: *)
(*Helper, prepares the association of connector labels:*)


logicConnectorLabels[inNames_List,outNames_List,labels_]:=With[
	{
		inOutNames=Join[inNames,outNames]
	},

	If[labels===Automatic,
		AssociationMap[ToString,inOutNames],
		Join[
			AssociationMap[Function[name,None],inOutNames],
			Association[If[labels===None,{},labels]]
		]
	]
]


(* ::Text:: *)
(*Helpers, prepare the list of vertical connector offsets, lengths of input connectors (for OR gates) and the length of input side extension for the given number of gate inputs (input side extensions are short lines that are drawn above and below the main gate shape in order to accommodate wider set of inputs if there are more than 3 inputs):*)


logicConnectorYOffsets[numOfConnectors_Integer,numOfConnectorsOnTheOtherSide_Integer]:=Module[
	{
		halfNum=Quotient[numOfConnectors,2],
		offsets
	},

	offsets=Reverse[Range[-halfNum,halfNum]];

	If[EvenQ[numOfConnectors],
		If[OddQ[numOfConnectorsOnTheOtherSide],
			Delete[offsets,halfNum+1],
			Delete[offsets,1]
		],
		offsets
	]
]

logicORInputConnectorLengths[numOfInputs_Integer]:=Module[
	{
		halfNum=Quotient[numOfInputs,2],
		lengths
	},

	lengths=Map[
		Function[n,
			Piecewise[
				{{1.81,n==-1},
				{1.99,n==0},
				{1.81,n==1}},
				1.18
			]
		],
		Reverse[Range[-halfNum,halfNum]]
	];

	If[EvenQ[numOfInputs],
		Delete[lengths,halfNum+1],
		lengths
	]
]

logicInputExtensionLength[numOfInputs_Integer]:=Quotient[Max[numOfInputs-2,0],2]


(* ::Text:: *)
(*Helpers, refactored common implementations for AND/OR gates and their variants.*)


logicInternalCoordinates[gate_]:=With[
	{
		name=gate[[1]],
		inNames=gate[[2]],
		outNames=gate[[3]]
	},

	Flatten[
		Map[
			Function[inName,{name[inName]["x"],name[inName]["y"]}],
			Join[inNames,outNames]
		]
	]
]

logicInternalLayout[gate_,width_Integer]:=Module[
	{
		name=gate[[1]],
		inNames=gate[[2]],
		outNames=gate[[3]],
		insLength,outsLength,
		inputOffsets,outputOffsets,
		xFirstInCoordName,xFirstOutCoordName,
		yAxisCoordName
	},

	insLength=Length[inNames];
	outsLength=Length[outNames];
	inputOffsets=logicConnectorYOffsets[insLength,outsLength];
	outputOffsets=logicConnectorYOffsets[outsLength,insLength];

	(* determine which connector (input or output) is the "vertical layout axis":
	if either inputs and outputs (or both) are odd, then either middle input or middle output is the axis,
	else (that is, if both inputs and outputs are even) either "upper-middle" input or output is the axis;
	this is already taken into account in inputOffsets and outputOffsets, as computed by logicConnectorYOffsets *)

	yAxisCoordName=If[OddQ[outsLength],
		name[outNames[[Quotient[outsLength,2]+1]]]["y"], (* TODO: Median? Middle? some function for extracting middle element *)
		If[OddQ[insLength],
			name[inNames[[Quotient[insLength,2]+1]]]["y"],
			name[inNames[[Quotient[insLength,2]]]]["y"]
		]
	];

	xFirstInCoordName=name[First[inNames]]["x"];
	xFirstOutCoordName=name[First[outNames]]["x"];

	Join[
		(* width between first input and first output *)
		{
			xFirstInCoordName+width==xFirstOutCoordName
		},

		(* all inputs vertically aligned *)
		Map[
			Function[in, (* in is an input name *)
				name[in]["x"]==xFirstInCoordName
			],
			Rest[inNames]
		],

		(* all outputs vertically aligned *)
		Map[
			Function[out, (* out is an output name *)
				name[out]["x"]==xFirstOutCoordName
			],
			Rest[outNames]
		],

		(* input vertical coordinates with regard to axis *)
		MapThread[
			Function[{inName,inYOffset},
				If[name[inName]["y"]=!=yAxisCoordName,
					name[inName]["y"]==yAxisCoordName+inYOffset,
					Nothing
				]
			],
			{inNames,inputOffsets}
		],

		(* output coordinates *)
		MapThread[
			Function[{outName,outYOffset}, (* out is a pair: {output-name, output-Y-offset} *)
				If[name[outName]["y"]=!=yAxisCoordName,
					name[outName]["y"]==yAxisCoordName+outYOffset,
					Nothing
				]
			],
			{outNames,outputOffsets}
		]
	]
]

logicBoundingBox[gate_]:=With[
	{
		name=gate[[1]],
		inNames=gate[[2]],
		outNames=gate[[3]]
	},

	(* all logic gates and blocks have inputs and outputs spaced in the same way,
	so it the side with larger number of lines is defining the extent of the bounding box *)
	If[Length[inNames]>=Length[outNames],
		{
			{name[Last[inNames]]["x"],name[Last[inNames]]["y"]-1},
			{name[First[outNames]]["x"],name[First[inNames]]["y"]+1}
		},
		{
			{name[Last[inNames]]["x"],name[Last[outNames]]["y"]-1},
			{name[First[outNames]]["x"],name[First[outNames]]["y"]+1}
		}
	]
]


(* ::Text:: *)
(*Fixed common value, radius of the circle denoting negation of the gate' s output :*)


logicCircleRadius=0.5; (* in grid units *)


logicNOT[name_,inName_,outName_,OptionsPattern[{labels->None,style->None,comment->None}]]:=
	name=logicNOTT[
		SymbolName[Unevaluated[name]],
		inName,
		outName,
		logicConnectorLabels[{inName},{outName},OptionValue[labels]], (* connector labels association *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[logicNOT,HoldFirst]

Prodiams`internalCoordinates[gate_logicNOTT]:=With[
	{
		name=gate[[1]],
		inName=gate[[2]],
		outName=gate[[3]]
	},

	{
		name[inName]["x"],name[inName]["y"],
		name[outName]["x"],name[outName]["y"]
	}
]

Prodiams`internalLayout[gate_logicNOTT]:=With[
	{
		name=gate[[1]],
		inName=gate[[2]],
		outName=gate[[3]]
	},

	{
		name[inName]["x"]+8==name[outName]["x"],
		name[inName]["y"]==name[outName]["y"]
	}
]

Prodiams`boundingBox[gate_logicNOTT]:=With[
	{
		name=gate[[1]],
		inName=gate[[2]],
		outName=gate[[3]]
	},

	{
		{name[inName]["x"],name[inName]["y"]-2},
		{name[outName]["x"],name[outName]["y"]+2}
	}
]

Prodiams`specificStyle[gate_logicNOTT]:=gate[[5]]

Prodiams`draw[gate_logicNOTT,layout_,style_]:=Module[
	{
		name=gate[[1]],
		inName=gate[[2]],
		outName=gate[[3]],
		labels=gate[[4]],
		comment=gate[[6]],
		circleRadius=logicCircleRadius,
		xIn,yIn,
		xOut,yOut,
		drawing
	},

	xIn=layout[name[inName]["x"]];
	yIn=layout[name[inName]["y"]];
	xOut=layout[name[outName]["x"]];
	yOut=layout[name[outName]["y"]];

	drawing=Flatten[{
		style[Prodiams`connectorLineStyle],
		Line[{{xIn,yIn},{xIn+2,yIn}}],
		Line[{{xOut-2+2*circleRadius,yOut},{xOut,yOut}}],
		style[Prodiams`lineStyle],
		Circle[{xOut-2+circleRadius,yOut},circleRadius],
		style[Prodiams`fillStyle],EdgeForm[style[Prodiams`edgeStyle]],
		Triangle[{{xIn+2,yIn+2},{xIn+2,yIn-2},{xOut-2,yOut}}],

		(* connector labels, if specified: *)
		If[labels[inName]=!=None,
			style[Prodiams`labelFunction][labels[inName],{xIn+1.5,yIn},{1,-1}],
			Nothing
		],
		If[labels[outName]=!=None,
			style[Prodiams`labelFunction][labels[outName],{xOut-0.5,yOut},{-1,-1}],
			Nothing
		]
	}];

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]

Prodiams`connector[gate_logicNOTT,conName_]:=With[
	{
		name=gate[[1]]
	},

	Prodiams`anchorT[name[conName]["x"],name[conName]["y"]]
]


logicAND[name_,inNames_List,outName_,OptionsPattern[{labels->None,outputNegated->False,style->None,comment->None}]]:=
	name=logicANDT[
		SymbolName[Unevaluated[name]],
		inNames,
		{outName},
		logicConnectorLabels[inNames,{outName},OptionValue[labels]], (* connector labels association *)
		OptionValue[outputNegated], (* True for NAND, False for AND *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[logicAND,HoldFirst]

Prodiams`internalCoordinates[gate_logicANDT]:=logicInternalCoordinates[gate];

Prodiams`internalLayout[gate_logicANDT]:=logicInternalLayout[gate,8];

Prodiams`boundingBox[gate_logicANDT]:=logicBoundingBox[gate];

Prodiams`specificStyle[gate_logicANDT]:=gate[[6]]

Prodiams`draw[gate_logicANDT,layout_,style_]:=Module[
	{
		name=gate[[1]],
		inNames=gate[[2]],
		outName=First[gate[[3]]],
		labels=gate[[4]],
		outputNegated=gate[[5]], (* True for NAND, False for AND *)
		comment=gate[[7]],
		circleRadius=logicCircleRadius,
		ins, (* association: input name \[Rule] coordinates {x,y} *)
		xOut,yOut,
		inputSideExtension,
		drawing
	},

	ins=AssociationMap[
		Function[inName,
			{layout[name[inName]["x"]],layout[name[inName]["y"]]}
		],
		inNames
	];
	inputSideExtension=logicInputExtensionLength[Length[inNames]];
	xOut=layout[name[outName]["x"]];
	yOut=layout[name[outName]["y"]];

	drawing=Flatten[{

		(* internal fill *)
		EdgeForm[style[Prodiams`fillStyle]],FaceForm[style[Prodiams`fillStyle]],
		Rectangle[{xOut-6,yOut-2},{xOut-4,yOut+2}],
		Disk[{xOut-4,yOut},{2,2},{-1.6,1.6}],

		(* edge *)
		style[Prodiams`edgeStyle],
		Line[{{xOut-4,yOut-2},{xOut-6,yOut-2},{xOut-6,yOut+2},{xOut-4,yOut+2}}],
		Circle[{xOut-4,yOut},{2,2},{-1.6,1.6}],

		(* input connectors *)
		style[Prodiams`connectorLineStyle],
		Map[
			Function[in,
				Line[{{in[[1]],in[[2]]},{in[[1]]+2,in[[2]]}}]
			],
			Values[ins]
		],

		(* input side extensions, if needed for more than 3 inputs *)
		If[inputSideExtension>0,
			{
				Line[{{xOut-6,yOut+2},{xOut-6,yOut+2+inputSideExtension}}],
				Line[{{xOut-6,yOut-2},{xOut-6,yOut-2-inputSideExtension}}]
			},
			Nothing
		],

		(* output connector and output negation circle, if needed *)
		If[outputNegated,
			{
				style[Prodiams`connectorLineStyle],
				Line[{{xOut-2+2*circleRadius,yOut},{xOut,yOut}}],
				style[Prodiams`lineStyle],
				Circle[{xOut-2+circleRadius,yOut},circleRadius]
			},
			{
				style[Prodiams`connectorLineStyle],
				Line[{{xOut-2,yOut},{xOut,yOut}}]
			}
		],

		(* input connector labels, if specified *)
		Map[
			Function[inName,
				If[labels[inName]=!=None,
					style[Prodiams`labelFunction][
						labels[inName], (* label text *)
						{ins[inName][[1]]+1.5,ins[inName][[2]]}, (* label coordinates *)
						{1,-1} (* label offsets *)
					],
					Nothing
				]
			],
			inNames
		],

		(* output connector label, if specified *)
		If[labels[outName]=!=None,
			style[Prodiams`labelFunction][labels[outName],{xOut-0.5,yOut},{-1,-1}],
			Nothing
		]
	}];

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]

Prodiams`connector[gate_logicANDT,conName_]:=With[
	{
		name=gate[[1]]
	},

	Prodiams`anchorT[name[conName]["x"],name[conName]["y"]]
]


logicNAND[name_,inNames_List,outName_,OptionsPattern[{labels->None,style->None,comment->None}]]:=
	logicAND[name,inNames,outName,labels->OptionValue[labels],outputNegated->True,style->OptionValue[style],comment->OptionValue[comment]]

SetAttributes[logicNAND,HoldFirst]


logicOR[name_,inNames_List,outName_,OptionsPattern[{labels->None,outputNegated->False,inputsExclusive->False,style->None,comment->None}]]:=
	name=logicORT[
		SymbolName[Unevaluated[name]],
		inNames,
		{outName},
		logicConnectorLabels[inNames,{outName},OptionValue[labels]], (* connector labels association *)
		OptionValue[outputNegated], (* True for NOR/XNOR, False for OR/XOR *)
		OptionValue[inputsExclusive], (* True for XOR/XNOR, False for OR/NOR *)
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[logicOR,HoldFirst]

Prodiams`internalCoordinates[gate_logicORT]:=logicInternalCoordinates[gate];

Prodiams`internalLayout[gate_logicORT]:=logicInternalLayout[gate,8];

Prodiams`boundingBox[gate_logicORT]:=logicBoundingBox[gate];

Prodiams`specificStyle[gate_logicORT]:=gate[[7]]

Prodiams`draw[gate_logicORT,layout_,style_]:=Module[
	{
		name=gate[[1]],
		inNames=gate[[2]],
		outName=First[gate[[3]]],
		labels=gate[[4]],
		outputNegated=gate[[5]], (* True for NOR/XNOR, False for OR/XOR *)
		inputsExclusive=gate[[6]], (* True for XOR/XNOR, False for OR/NOR *)
		comment=gate[[8]],
		circleRadius=logicCircleRadius,
		ins, (* association: input name \[Rule] coordinates for Line {{x,y},{x,y}} *)
		xOut,yOut,
		inputSideExtension,
		drawing
	},

	ins=Association[
		MapThread[
			Function[{inName,conLength},
				inName->
					{
						{layout[name[inName]["x"]],layout[name[inName]["y"]]},
						{layout[name[inName]["x"]]+conLength,layout[name[inName]["y"]]}
					}
			],
			{inNames,logicORInputConnectorLengths[Length[inNames]]}
		]
	];
	inputSideExtension=logicInputExtensionLength[Length[inNames]];
	xOut=layout[name[outName]["x"]];
	yOut=layout[name[outName]["y"]];

	drawing=Flatten[{
		(* internal fill *)
		EdgeForm[style[Prodiams`fillStyle]],FaceForm[style[Prodiams`fillStyle]],
		Rectangle[{xOut-6.82,yOut-2},{xOut-5,yOut+2}],
		Disk[{xOut-5,yOut-1},{3,3 (* 1 + height/2 *)},{0.34,1.6}],
		Disk[{xOut-5,yOut+1},{3,3 (* 1 + height/2 *)},{-0.34,-1.6}],
		EdgeForm[White],FaceForm[White],
		Disk[{xOut-8.84,yOut},{2.84,2.84 (* 0.84 + height/2 *)},{-0.78,0.78}],

		(* edge *)
		style[Prodiams`edgeStyle],
		Line[{{xOut-6.82,yOut-2},{xOut-5,yOut-2}}],
		Line[{{xOut-6.82,yOut+2},{xOut-5,yOut+2}}],
		Circle[{xOut-5,yOut-1},{3,3 (* 1 + height/2 *)},{0.34,1.58}],
		Circle[{xOut-5,yOut+1},{3,3 (* 1 + height/2 *)},{-0.34,-1.58}],
		Circle[{xOut-8.84,yOut},{2.84,2.84 (* 0.84 + height/2 *)},{-0.78,0.78}],
		If[inputsExclusive,
			Circle[{xOut-9.52,yOut},{2.84,2.84 (* 0.84 + height/2 *)},{-0.78,0.78}],
			Nothing
		],

		(* input connectors *)
		style[Prodiams`connectorLineStyle],
		Map[Line,Values[ins]],

		(* input side extensions, if needed for more than 3 inputs *)
		If[inputSideExtension>0,
			{
				style[Prodiams`lineStyle],
				Line[{{xOut-6.82,yOut+2},{xOut-6.82,yOut+2+inputSideExtension}}],
				Line[{{xOut-6.82,yOut-2},{xOut-6.82,yOut-2-inputSideExtension}}],
				If[inputsExclusive,
					{
						Line[{{xOut-7.50,yOut+2},{xOut-7.50,yOut+2+inputSideExtension}}],
						Line[{{xOut-7.50,yOut-2},{xOut-7.50,yOut-2-inputSideExtension}}]
					},
					Nothing
				]
			},
			Nothing
		],

		(* output connector and output negation circle, if needed *)
		If[outputNegated,
			{
				style[Prodiams`connectorLineStyle],
				Line[{{xOut-2.16+2*circleRadius,yOut},{xOut,yOut}}],
				style[Prodiams`lineStyle],
				Circle[{xOut-2.16+circleRadius,yOut},circleRadius]
			},
			{
				style[Prodiams`connectorLineStyle],
				Line[{{xOut-2.16,yOut},{xOut,yOut}}]
			}
		],

		(* input connector labels, if specified *)
		Map[
			Function[inName,
				If[labels[inName]=!=None,
					style[Prodiams`labelFunction][
						labels[inName], (* label text *)
						{ins[inName][[1,1]]+If[inputsExclusive,0,1],ins[inName][[1,2]]}, (* label coordinates *)
						{1,-1} (* label offsets *)
					],
					Nothing
				]
			],
			inNames
		],

		(* output connector label, if specified *)
		If[labels[outName]=!=None,
			style[Prodiams`labelFunction][labels[outName],{xOut-0.5,yOut},{-1,-1}],
			Nothing
		]
	}];

	(* comment, if specified: *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]

Prodiams`connector[gate_logicORT,conName_]:=With[
	{
		name=gate[[1]]
	},

	Prodiams`anchorT[name[conName]["x"],name[conName]["y"]]
]


logicNOR[name_,inNames_List,outName_,OptionsPattern[{labels->None,style->None,comment->None}]]:=
	logicOR[name,inNames,outName,labels->OptionValue[labels],outputNegated->True,inputsExclusive->False,style->OptionValue[style],comment->OptionValue[comment]];

SetAttributes[logicNOR,HoldFirst]


logicXOR[name_,inNames_List,outName_,OptionsPattern[{labels->None,style->None,comment->None}]]:=
	logicOR[name,inNames,outName,labels->OptionValue[labels],outputNegated->False,inputsExclusive->True,style->OptionValue[style],comment->OptionValue[comment]];

SetAttributes[logicXOR,HoldFirst]


logicXNOR[name_,inNames_List,outName_,OptionsPattern[{labels->None,style->None,comment->None}]]:=
	logicOR[name,inNames,outName,labels->OptionValue[labels],outputNegated->True,inputsExclusive->True,style->OptionValue[style],comment->OptionValue[comment]];

SetAttributes[logicXNOR,HoldFirst]


logicBlock[name_,label_String,inNames_List,outNames_List,OptionsPattern[{width->16,connectorLabels->Automatic,style->None,comment->None}]]:=
	name=logicBlockT[
		SymbolName[Unevaluated[name]],
		inNames,
		outNames,
		logicConnectorLabels[inNames,outNames,OptionValue[connectorLabels]], (* connector labels association *)
		OptionValue[width],
		label,
		OptionValue[style],
		OptionValue[comment]
	]

SetAttributes[logicBlock,HoldFirst]

Prodiams`internalCoordinates[b_logicBlockT]:=logicInternalCoordinates[b];

Prodiams`internalLayout[b_logicBlockT]:=With[
	{
		width=b[[5]]
	},

	logicInternalLayout[b,width]
]

Prodiams`boundingBox[b_logicBlockT]:=logicBoundingBox[b];

Prodiams`specificStyle[b_logicBlockT]:=b[[7]]

Prodiams`draw[b_logicBlockT,layout_,style_]:=Module[
	{
		name=b[[1]],
		inNames=b[[2]],
		outNames=b[[3]],
		connectorLabels=b[[4]],
		label=b[[6]], (* True for NAND, False for AND *)
		comment=b[[8]],
		ins, (* association: input name \[Rule] coordinates {x,y} *)
		outs, (* association: output name \[Rule] coordinates {x,y} *)
		xCenter, yCenter,
		drawing
	},

	ins=AssociationMap[
		Function[inName,
			{layout[name[inName]["x"]],layout[name[inName]["y"]]}
		],
		inNames
	];
	outs=AssociationMap[
		Function[outName,
			{layout[name[outName]["x"]],layout[name[outName]["y"]]}
		],
		outNames
	];

	(* center point (for placing the main label) is in the middle of inputs and outputs: *)
	xCenter=Mean[{layout[name[First[inNames]]["x"]],layout[name[First[outNames]]["x"]]}];
	yCenter=Mean[Values[ins][[All,2]]];

	drawing=Flatten[{

		(* internal fill *)
		style[Prodiams`fillStyle],EdgeForm[style[Prodiams`edgeStyle]],
		Rectangle[{Last[Values[ins]][[1]]+2,Min[Last[Values[ins]][[2]],Last[Values[outs]][[2]]]-1},
		{First[Values[outs]][[1]]-2,Max[First[Values[ins]][[2]],First[Values[outs]][[2]]]+1}],

		(* input connectors *)
		style[Prodiams`connectorLineStyle],
		Map[
			Function[in,
				Line[{{in[[1]],in[[2]]},{in[[1]]+2,in[[2]]}}]
			],
			Values[ins]
		],
		Map[
			Function[out,
				Line[{{out[[1]]-2,out[[2]]},{out[[1]],out[[2]]}}]
			],
			Values[outs]
		],

		(* main label *)
		style[Prodiams`labelFunction][
			label,
			{xCenter,yCenter},
			{0,0}
		],

		(* input and output connector labels, if specified *)
		Map[
			Function[inName,
				If[connectorLabels[inName]=!=None,
					style[Prodiams`labelFunction][
						connectorLabels[inName], (* label text *)
						{ins[inName][[1]]+1.5,ins[inName][[2]]}, (* label coordinates *)
						{1,-1} (* label offsets *)
					],
					Nothing
				]
			],
			inNames
		],
		Map[
			Function[outName,
				If[connectorLabels[outName]=!=None,
					style[Prodiams`labelFunction][
						connectorLabels[outName], (* label text *)
						{outs[outName][[1]]-1.5,outs[outName][[2]]}, (* label coordinates *)
						{-1,-1} (* label offsets *)
					],
					Nothing
				]
			],
			outNames
		]
	}];

	(* comment, if specified *)
	If[comment=!=None,
		Tooltip[drawing,comment],
		drawing
	]
]

Prodiams`connector[b_logicBlockT,conName_]:=With[
	{
		name=b[[1]]
	},

	Prodiams`anchorT[name[conName]["x"],name[conName]["y"]]
]


(* ::Text:: *)
(*Diagram-generation functions.*)


(* ::Text:: *)
(*Determines whether the index is a leaf (has no children):*)


expressionLeafQ[idx_]:=Last[idx]!=0


(* ::Text:: *)
(*Deconstructs a given expression into an association partIndex->symbol, where leaf symbols are separately tagged:*)


deconstructExpression[expr_]:=Association[
	If[Head[expr]===Symbol,
		{1}->expressionLeafT[expr],
		Rule@@@Reap[
			MapIndexed[
				Function[{part,idx},
					If[Head[part]===Symbol,
						If[expressionLeafQ[idx],
							(* leaf symbol (variable) *)
							Sow[{idx,expressionLeafT[part]}],
							(* node (function call) *)
							Sow[{idx,part}]
						]
					]
				],
				expr,
				Infinity,
				Heads->True
			]
		][[2,1]]
	]
]


(* ::Text:: *)
(*Creates, from the list of part indices, a set of rules for expression's tree graph:*)


treeFromIndices[indices_]:=If[Length[indices]>1,
	Map[
		Function[idx,
			With[
				{
					most=Most[idx],
					from=Join[Most[idx],{0}]
				},
				
				If[expressionLeafQ[idx],
					(* idx is a leaf *)
					from->idx,
					
					(* idx is a non-leaf node *)
					If[most!={},
						Join[Most[most],{0}]->idx,
						Nothing
					]
				]
			]
		],
		indices
	],
	{}
]


(* ::Text:: *)
(*Orders indices in their lexicographical order (standard Order places shorter argument before the longer one, which is not the natural tree order for indices:*)


indicesInOrder[idx1_,idx2_]:=Module[
	{
		len1=Length[idx1],
		len2=Length[idx2],
		ord
	},
	
	ord=If[len1<len2,
		Order[idx1,Take[idx2,len1]],
		If[len1>len2,
			Order[Take[idx1,len2],idx2],
			Order[idx1,idx2]
		]
	];

	ord>=0
]


(* ::Text:: *)
(*Creates a list of rules for ordering nodes across the same level:*)


orderAtSameLevel[indices_]:=Flatten[
	Map[
		Function[level,
			Rule@@@Partition[level,2,1]
		],
		GatherBy[indices,
			Function[idx,
				If[expressionLeafQ[idx],Length[idx],Length[idx] - 1]
			]
		]
	]
]


(* ::Text:: *)
(*Assuming that indices refer to neighbouring leafs, the order is created for shapes at the same level - the level is that of the shorter index;*)
(*this allows to create a necessary space between leaves for gates (nodes) higher in the tree, for example in And[A,Or[Not[B],C]], where A would conflict with Not without this tratment:*)


makeOrderForLeafPair[idx1_,idx2_]:=With[
	{
		len1=Length[idx1],
		len2=Length[idx2]
	},
	
	If[len1<len2,
		idx1->Join[Most[idx2],{0}],
		If[len1>len2,
			Join[Most[idx1],{0}]->idx2,
			idx1->idx2
		]
	]
]


(* ::Text:: *)
(*Creates a list of rules for ordering all leaves, taking into account different branch lengths:*)


leafOrder[indices_]:=makeOrderForLeafPair@@@Partition[
	Sort[
		Select[indices,expressionLeafQ],
		indicesInOrder
	],
	2,1
]


(* ::Text:: *)
(*Returns the list of children for the given parent:*)


children[indices_,parent_]:=With[
	{
		pLen=Length[parent],
		pMost=Most[parent]
	},
	
	If[Last[parent]==0,
		Select[indices,
			Function[idx,
				With[
					{
						idxLen=Length[idx],
						idxMost=Most[idx],
						idxLast=Last[idx]
					},
					
					(idxLen==pLen&&idxMost==pMost&&idxLast!=0)||
					(idxLen==pLen+1&&Most[idxMost]==pMost&&idxLast==0)
				]
			]
		],
		
		(* parent is a leaf already *)
		{}
	]
]


(* ::Text:: *)
(*Creates order for non-leaves (non-leaf is placed in the middle of its middle branches),*)
(* result is {{p->{ch1,ch2},...},{p->ch,...}}, first sublist is ordering in the middle of two middle children (if the number of children is even) and second sublist is centering (if number of children is odd, then the middle child is chosen), any given parent has its entry in only one of these two lists:*)


orderParentsBetweenChildren[indices_]:=Module[
	{
		between={},
		middle={},
		ch,len
	},

	Scan[
		Function[parent,
			ch=children[indices,parent];
			len=Length[ch];
			
			If[len>0,
				If[EvenQ[len],
					AppendTo[between,parent->{ch[[len/2]],ch[[len/2+1]]}],
					
					(* odd number of children *)
					AppendTo[middle,parent->ch[[(len+1)/2]]]
				]
			]
		],
		indices
	];

	{between,middle}
]


(* ::Text:: *)
(*Makes the name of input, based on its order number:*)


inputName[i_]:="In"~~ToString[i]


(* ::Text:: *)
(*Returns the list of input names appropriate for the arity of the parent:*)


inputNames[indices_,parent_]:=With[
	{
		ch=children[indices,parent]
	},
	
	Table[inputName[i],{i,Length[ch]}]
]


(* ::Text:: *)
(*Returns the name of parent input based on the child index:*)


parentInputNameFromChildIndex[childIdx_]:=With[
	{
		last=Last[childIdx]
	},
	
	If[last!=0,
		inputName[last],
		inputName[Last[Most[childIdx]]]
	]
]


(* ::Text:: *)
(*Generates symbols for all nodes of the deconstructed expression - gates for non-leaves and inPorts for leaves:*)


makeDiagramShapes[decExpr_]:=With[
	{
		indices=Keys[decExpr]
	},
	
	Association[
		Join[
			(* gates and inports *)
			Map[
				Function[idx,
					With[
						{
							node=decExpr[idx]
						},
						
						idx->Switch[node,
							Not,ProdiamsLogic`logicNOT[Evaluate[Unique["not"]],inputName[1],"Out"],
							
							And,ProdiamsLogic`logicAND[Evaluate[Unique["and"]],inputNames[indices,idx],"Out"],
							
							Nand,ProdiamsLogic`logicNAND[Evaluate[Unique["nand"]],inputNames[indices,idx],"Out"],
							
							Or,ProdiamsLogic`logicOR[Evaluate[Unique["or"]],inputNames[indices,idx],"Out"],
							
							Nor,ProdiamsLogic`logicNOR[Evaluate[Unique["nor"]],inputNames[indices,idx],"Out"],
							
							Xor,ProdiamsLogic`logicXOR[Evaluate[Unique["xor"]],inputNames[indices,idx],"Out"],
							
							Xnor,ProdiamsLogic`logicXNOR[Evaluate[Unique["xnor"]],inputNames[indices,idx],"Out"],
							
							If,ProdiamsLogic`logicBlock[Evaluate[Unique["if"]],"If/Then/Else",inputNames[indices,idx],{"Out"},width->8,connectorLabels->{inputName[1]->"if",inputName[2]->"then",inputName[3]->"else","Out"->None}],
							
							Implies,ProdiamsLogic`logicBlock[Evaluate[Unique["implies"]],"\[Implies]",inputNames[indices,idx],{"Out"},width->8,connectorLabels->{inputName[1]->"pre",inputName[2]->"con","Out"->None}],
							
							expressionLeafT_,Prodiams`inPort[Evaluate[Unique["in"]],ToString[node[[1]]]]
						]
					]
				],
				indices
			],
			
			(* finally, add the result outport *)
			{"result"->Prodiams`outPort[Evaluate[Unique["out"]],"Result"]}
		]
	]
]


(* ::Text:: *)
(*Makes the connection lines based on the parent connectivity tree;*)
(*the key in the result association is the connection rule from the argument tree:*)


makeConnectorLines[gateTree_]:=Map[
	Function[pair,
		With[
			{
				from=pair[[1]],
				to=pair[[2]]
			},
			
			pair->Prodiams`line[Evaluate[Unique["line"]],{"A","B","C","D"}]
		]
	],
	gateTree
]


(* ::Text:: *)
(*Converts the list of rules between indices (the rules giving some ordering)*)
(*into shape-relation calls, like after, above, etc.:*)


shapeRelationsFromIndiceRules[shapes_,indiceRules_,relation_]:=Map[
	Function[pair,
		With[
			{
				parent=shapes[pair[[1]]],
				child=shapes[pair[[2]]]
			},
			
			relation[parent,child]
		]
	],
	indiceRules
]


(* ::Text:: *)
(*Returns the connection point for the output of the given shape,*)
(*whether it is leaf (inPort, in which case the whole inPort is its own connection point) or a node (which has a proper "Out" connection point:*)


shapeOutput[shapes_,idx_]:=If[expressionLeafQ[idx],
	shapes[idx],
	Prodiams`connector[shapes[idx],"Out"]
]


(* ::Text:: *)
(*Returns the vertical coordinate of the shape, whether it is leaf (inPort) or a node (more complex shape):*)


shapeCoordinateY[shapes_,idx_]:=If[expressionLeafQ[idx],
	shapes[idx][[2]],
	Prodiams`connector[shapes[idx],"Out"][[2]]
]


(* ::Text:: *)
(*Creates actual constraints for ordering parents between children, as determined by orderParentsBetweenChildren:*)


orderParentsBetweenChildrenConstr[shapes_,orderPair_]:=Module[
	{
		parentIdx,
		childIdx,child1Idx,child2Idx,
		parentOuty,
		childOuty,child1Outy,child2Outy
	},
	
	Join[
		(* order for parents between two of their children *)
		Map[
			Function[middleOrder,
				parentIdx=middleOrder[[1]];
				child1Idx=middleOrder[[2,1]];
				child2Idx=middleOrder[[2,2]];
				
				(* performance optimization: avoid Min/Max that are internally used by Prodiams`extent *)
				(*
				middle[
					shapes[parentIdx],
					extent[connector[shapes[child1Idx],"Out"],connector[shapes[child2Idx],"Out"]]
				]
				*)
				(* instead, operate directly on vertical values for the Out connectors of all involved shapes *)
				parentOuty=shapeCoordinateY[shapes,parentIdx];
				child1Outy=shapeCoordinateY[shapes,child1Idx];
				child2Outy=shapeCoordinateY[shapes,child2Idx];
				
				(* ordering strictly in the middle is not necessary:
				{child1Outy+child2Outy\[Equal]2*parentOuty}
				*)
				{child1Outy>parentOuty>child2Outy}
			],
			orderPair[[1]]
		],
		
		(* order for parents that have a single middle child *)
		Map[
			Function[middleOrder,
				parentIdx=middleOrder[[1]];
				childIdx=middleOrder[[2]];
				
				parentOuty=shapeCoordinateY[shapes,parentIdx];
				childOuty=shapeCoordinateY[shapes,childIdx];
				
				{parentOuty==childOuty}
			],
			orderPair[[2]]
		]
	]
]


(* ::Text:: *)
(*Attaches the result outPort to the root gate:*)


attachResultPort[shapes_,indices_]:=If[Length[indices]==1,
	{Prodiams`pin[shapes["result"],shapes[{1}]]},
	{Prodiams`pin[shapes["result"],Prodiams`connector[shapes[{0}],"Out"]]}
]


(* ::Text:: *)
(*Connects connector lines to their respective elements (leaves or outputs or lower gates on the left, input ports of higher gate on the right) and prepares the knee at 90deg (but without determining the horizontal placement of the knee):*)


routeConnectorLines[shapes_,lines_]:=Map[
	Function[lineSpec,
		With[
			{
				(* lineSpec has the form (parent\[Rule]child)\[Rule]lineObject *)
				parentIdx=lineSpec[[1,1]],
				childIdx=lineSpec[[1,2]],
				line=lineSpec[[2]]
			},
			
			{
				Prodiams`pin[shapeOutput[shapes,childIdx],Prodiams`linePoint[line,"A"]],
				Prodiams`after[Prodiams`linePoint[line,"B"],Prodiams`linePoint[line,"A"]],
				Prodiams`middle[Prodiams`linePoint[line,"B"],Prodiams`linePoint[line,"A"]],
				Prodiams`center[Prodiams`linePoint[line,"B"],Prodiams`linePoint[line,"C"]],
				Prodiams`after[Prodiams`linePoint[line,"D"],Prodiams`linePoint[line,"C"]],
				Prodiams`middle[Prodiams`linePoint[line,"D"],Prodiams`linePoint[line,"C"]],
				Prodiams`pin[
					Prodiams`connector[shapes[parentIdx],
					parentInputNameFromChildIndex[childIdx]],Prodiams`linePoint[line,"D"]
				]
			}
		]
	],
	lines
]


(* ::Text:: *)
(*Orders line knees in the given list, so that vertical segments (from "B" to "C") do not overlap,*)
(*this is done by partitioning the list into pairs and arranging those pairs with the after layout directive;*)
(*this function should be called for lines belonging to the single parent:*)


orderLineKneesForSingleParent[lines_]:=Module[
	{
		len=Length[lines],
		lineOrderedPairs={},
		singleLines={}
	},
	
	If[len>=4,
		lineOrderedPairs=If[EvenQ[len],
			Join[
				Partition[lines[[1;;len/2]],2,1],
				Map[Reverse,Partition[lines[[len/2+1;;]],2,1]]
			],
			Join[
				Partition[lines[[1;;(len-1)/2]],2,1],
				Map[Reverse,Partition[lines[[(len+1)/2+1;;]],2,1]]
			]
		],
		singleLines=lines
	];
	
	Join[
		(* if there are at least 4 inputs, route lines so that they do not cross *)
		Map[
			Function[linePair,
				With[
					{
						furtherLine=linePair[[1]],
						closerLine=linePair[[2]]
					},
					
					Prodiams`after[
						Prodiams`linePoint[furtherLine,"B"],
						Prodiams`linePoint[closerLine,"B"]
					]
				]
			],
			lineOrderedPairs
		],
		
		(* if there are less than 4 inputs, route them so that their knees are closer to the right, to avoid conflicts with any shapes in the left part of the diagram *)
		Map[
			Function[line,
				Prodiams`after[
					Prodiams`linePoint[line,"C"],
					Prodiams`linePoint[line,"D"],
					offset->-1
				]
			],
			singleLines
		],
		
		(* align left line ends (which means - elements one layer down) vertically *)
		{
			Equal@@Map[
				Function[line,
					Prodiams`linePoint[line,"A"][[1]]
				],
				lines
			]
		}
	]
]

orderLineKnees[lines_]:=Module[
	{
		linesGroupedByParent=GatherBy[lines,Function[line,line[[1,1]]]]
	},
	
	Flatten[
		Map[
			Function[lineGroupForSingleParent,
				(* lineGroupForSingleParent has the form {(parent\[Rule]child)\[Rule]line,...} *)
				orderLineKneesForSingleParent[lineGroupForSingleParent[[All,2]]]
			],
			linesGroupedByParent
		]
	]
]


(* ::Text:: *)
(*Main function:*)


logicDiagram[expr_,OptionsPattern[{style->None,debug->False}]]:=Module[
	{
		dec,
		indices,
		tree,
		shapes,
		connectorLines,
		leavesOrdered,
		sameLevelOrder,
		parentMidChildOrderPair,parentMidChildOrder,
		resultAttached,
		connectorLinesRouted,
		orderedKnees,
		allConstraints,
		printDebug=OptionValue[debug]
	},
	
	(* deconstruct expression into association of indexed components *)
	dec=deconstructExpression[expr];
	indices=Sort[Keys[dec],indicesInOrder];
	tree=treeFromIndices[indices];
	
	If[printDebug,
		Print["deconstructed: ",dec];
		Print["indices: ",indices];
		Print["tree: ",tree]
	];
	
	(* generate shape symbols (both nodes and connecting lines) *)
	shapes=makeDiagramShapes[dec];
	connectorLines=makeConnectorLines[tree];
	
	If[printDebug,
		Print["shapes: ",shapes];
		Print["lines: ",connectorLines]
	];
	
	(* generate relations between diagram nodes *)
	
	(* parents are further on the right (after)  than their children,
	but this relation is forced by connector lines, which connect some output on the left with some input on the right *)
	
	(* all inputs are one above the other (and since their gates downstream are positioned between them, they will not conflict, either *)
	leavesOrdered=shapeRelationsFromIndiceRules[
		shapes,
		leafOrder[indices],
		Prodiams`above
	];
	
	If[printDebug,
		Print["leaves ordered: ",leavesOrdered]
	];
	
	(* nodes at the same level of expression are ordered vertically (one above the other) *)
	sameLevelOrder=shapeRelationsFromIndiceRules[
		shapes,
		orderAtSameLevel[indices],
		Prodiams`above
	];
	
	If[printDebug,
		Print["sameLevelOrder: ",sameLevelOrder]
	];
	
	(* align parents vertically with their middle child or children *)
	parentMidChildOrderPair=orderParentsBetweenChildren[indices];
	parentMidChildOrder=orderParentsBetweenChildrenConstr[shapes,parentMidChildOrderPair];
	
	If[printDebug,
		Print["parentMidChildOrderPair: ",parentMidChildOrderPair];
		Print["parentMidChildOrder: ",parentMidChildOrder]
	];
	
	(* result is directly attached to the top-level gate's output *)
	resultAttached=attachResultPort[shapes,indices];
	
	(* connector lines have their ends attached to parent and child gates and their middle points are routed at 90 deg angles *)
	connectorLinesRouted=routeConnectorLines[shapes,connectorLines];
	
	If[printDebug,
		Print["connectorLinesRouted: ",connectorLinesRouted]
	];
	
	(* for gates with multiple inputs (>4) the connector lines have knees (middle points) ordered so that their vertical segments do not overlap *)
	orderedKnees=orderLineKnees[connectorLines];
	
	If[printDebug,
		Print["orderedKnees: ",orderedKnees]
	];
	
	allConstraints={
		leavesOrdered,
		sameLevelOrder,
		parentMidChildOrder,
		resultAttached,
		connectorLinesRouted,
		orderedKnees
	};
	
	(*allConstraints={};*)
	If[printDebug,
		Print["allConstraints: ",allConstraints]
	];

	Prodiams`diagram[Join[Values[shapes],Values[connectorLines]],
		allConstraints,
		style->OptionValue[style]
	]
]


End[];


EndPackage[];
